#!/usr/bin/perl -w

#
# vCenter Discovery and Information Gathering
#
# Gather information from vCenter, Hosts and VMs.
# Supported operations are:
#   dclist - list of datacenters configured in the vcenter
#   hostlist - list of hosts in the vcenter
#   vmlist - list of VMs (grouped by hosts) in the vcenter
#   hostinfo - ESXi host information
#   maintmode - Check maintenance mode on a vSphere host
#   vminfo - VM information
#   vdiskinfo - VM disk info
#
use strict;
use warnings;
use Switch;

use VMware::VIRuntime;
use VMware::VILib;
use VMware::VICredStore;

my %opts = (
	    'vmname' => {
		type => "=s",
		help => "VM name",
		required => 0,
	    },
	    'hostname' => {
		type => "=s",
		help => "Host name",
		required => 0,
	    },
	    'datacenter' => {
		type => "=s",
		help => "Name of the DataCentre",
		required => 0,
	    },
	    'operation' => {
		type => "=s",
		variable => "operation",
		help => "Operation to do: dclist, hostlist, vmlist, hostinfo, maintmode, vminfo, vdiskinfo",
		required => 1,
	    }
);

Opts::add_options(%opts);
Opts::parse();
Opts::validate();

Util::connect();

sub listDC()
{
	my $vCenter_service_instance = Vim::get_service_instance(properties => ["content"]);
	my $vCenter_version = $vCenter_service_instance->content->about->version;

	if (!($vCenter_version =~ /5\.0/)) {
		print "Error: vCenter version ".$vCenter_version. " not supported";
		return;
	}

    my $datacenter_view = Vim::find_entity_views(view_type => 'Datacenter',
						 properties => [ "name" ] );
    foreach (@$datacenter_view) 
    {
	print $_->name ."\n";
    }
}


sub get_entities {
   my %args = @_;
   my $view_type = $args{view_type};
   my $obj = $args{obj};

   my $sc = Vim::get_service_content();
   my $service = Vim::get_vim_service();

   my $property_spec = PropertySpec->new(all => 0,
                                         type => $view_type->get_backing_type());
   my $property_filter_spec = $view_type->get_search_filter_spec($obj,[$property_spec]);
   my $obj_contents = $service->RetrieveProperties(_this => $sc->propertyCollector,
                                                   specSet => $property_filter_spec);
   my $result = Util::check_fault($obj_contents);
   return $result;
}

#
# List all the hosts managed by a vcenter
#
sub listHosts($)
{
    my ($dcname) = @_;

    my $hostlist;
    my $datacenter;

    if (defined $dcname) {
	my $datacenter_views = Vim::find_entity_views (view_type => 'Datacenter',
						       filter => {name => $dcname});
	die "Datacenter '$dcname' not found" unless (@$datacenter_views);

	foreach(@$datacenter_views) {
	    $datacenter = $_;
	}

	my $result = get_entities(view_type => 'HostSystem',
				  obj => $datacenter);
	foreach (@$result) {
	    my $obj_content = $_;
	    my $mob = $obj_content->obj;
	    my $obj = Vim::get_view(mo_ref=>$mob,
				    properties => [ "name" ] );
	    print "Host : " . $obj->name . "\n";
	}

    } else {
	eval {
	    $hostlist = Vim::find_entity_views(view_type => 'HostSystem',
					       properties => [ "summary" ] );
	};
	die $@ if ($@);
	foreach (@$hostlist) {
	    print "Host : " . $_->summary->config->name . "\n";
	}
    }
}

# Get sfcbd-watchdog status
# Args:
#	$host - Host system,
#
# Returns:
#	nothing
sub serviceGetSfcbdWatchdogStatus {
	my ($hostname) = @_;
	my $sfcbdWatchdogStatus = PrimType->new(0, 'boolean');
	my $hostServiceSystem;
	my $serviceSystem;

	my $host = Vim::find_entity_view(
                view_type => 'HostSystem',
                filter => { name => $hostname },
                properties => ["configManager"],
                );

	eval {
		$hostServiceSystem = Vim::get_view(mo_ref => $host->configManager->serviceSystem);
	};
	if ($@) {
		die "Error: Could not connect to host";
	}
	
	eval { 
		$hostServiceSystem->RefreshServices();
	};
	if ($@) {
		die "Error: Could not refresh services";
	}
	
	eval {
		$serviceSystem = Vim::get_view(mo_ref => $host->configManager->serviceSystem);
	};
	if ($@) {
		die "Error: Could not connect to host";
	}

	my $services = $serviceSystem->serviceInfo->service;

	foreach my $service (@$services) {
		my $key = $service->key;
		if ($key eq "sfcbd-watchdog") {
			if ($service->running eq 1) {
				print "SFCBD Status:\tStarted\n";
			} else {
				print "SFCBD Status:\tStopped\n";
			}
			return;
		}
	}
}	

#
# Restart the sfcbd-watchdog
# Args:
#   $host - Host system,
#
# Returns:
#   nothing
sub serviceSfcbdWatchdogRestart {
	my ($hostname) = @_;

	my $host = Vim::find_entity_view(
				view_type => 'HostSystem',
				filter => { name => $hostname },
				properties => ["configManager"],
				);

    my $hostServiceSystem = Vim::get_view(mo_ref => $host->configManager->serviceSystem);
    eval {
        $hostServiceSystem->RestartService(id => "sfcbd-watchdog");
    };
    if ($@) {
        die "Error: Could not restart the sfcbd-watchdog service";
    }
    print "Sfcbd-watchdog restarted\n";
}

#
# List all the Hosts and VMs managed by a vcenter
#
sub listHostsVMs($)
{
    my ($dcname) = @_;

    my $hostlist;
    my $datacenter;

    if (defined $dcname) {
	my $datacenter_views = Vim::find_entity_views (view_type => 'Datacenter',
						       filter => {name => $dcname});
	die "Datacenter '$dcname' not found" unless (@$datacenter_views);

	foreach(@$datacenter_views) {
	    $datacenter = $_;
	}

	my $result = get_entities(view_type => 'HostSystem',
				  obj => $datacenter);
	foreach (@$result) {
	    my $obj_content = $_;
	    my $mob = $obj_content->obj;
	    my $obj = Vim::get_view(mo_ref=>$mob,
				    properties => [ "name", "vm" ] );
	    print "Host : " . $obj->name . "\n";
	    my $vms;
	    eval {
		$vms =  Vim::get_views(mo_ref_array => $obj->vm,
				       properties => [ "summary" ]);
	    };
	    die $@ if ($@);
	    foreach (@$vms) {
		print "   VM : " . $_->summary->config->name . "\n";
	    }
	}

    } else {

	eval {
	    $hostlist = Vim::find_entity_views(view_type => 'HostSystem',
					       properties => [ "summary", "vm" ] );
	};
	die $@ if ($@);
	foreach (@$hostlist) {
	    print "Host : " . $_->summary->config->name . "\n";
	    my $vms;
	    eval {
		$vms =  Vim::get_views(mo_ref_array => $_->vm,
				       properties => [ "summary" ]);
	    };
	    die $@ if ($@);
	    foreach (@$vms) {
		print "   VM : " . $_->summary->config->name . "\n";
	    }
	}
    }
}

#
# Check to see if host is in maintenance mode
sub getMaintenanceMode($)
{
    my ($hostname) = @_;
    
    my $hostlist;
    die "No hostname specified" unless ($hostname);
    eval {
	$hostlist = Vim::find_entity_views(view_type => 'HostSystem',
					   filter => { name => $hostname } ,
					   properties => [ "runtime" ] );
    };
    die $@ if ($@);
    foreach (@$hostlist) {
	my $mode = $_->runtime->inMaintenanceMode;
	print "Maintenance mode: $mode\n";
    }
}

#
# Print information about the host
#
sub hostInfo($)
{
    my ($hostname) = @_;

    my $hostlist;
    die "No hostname specified" unless ($hostname);

    my %host_status = (
		       'gray' => 'The status is unknown',
		       'green' => 'The entity is OK',
		       'red' => 'The entity definitely has a problem',
		       'yellow' => 'The entity might have a problem',
		       );

    eval {
	$hostlist = Vim::find_entity_views(view_type => 'HostSystem',
					   filter => { name => $hostname } ,
					   properties => [ "name", "summary", "runtime", "hardware" ] );
    };
    die $@ if ($@);
    foreach (@$hostlist) {
	my $status = $_->summary->overallStatus->val;
	my $host_view = $_;

	print "Host Name:\t" . $host_view->name . "\n";
	print "Host Status:\t" . $host_status{$status} . "\n";

	if (defined ($host_view->runtime->bootTime)) {
	    print "BootTime:\t" . $host_view->runtime->bootTime . "\n";
	} else {
	    print "BootTime:\tNot Known\n";
	}
	if (defined ($host_view->summary->hardware)) {
	    print "Cpu Model:\t" . $host_view->summary->hardware->cpuModel . "\n";
	} else {
	    print "Cpu Model:\tNot Known\n";
	}
	if (defined ($host_view->hardware)) {
	    print "Cpu Speed:\t" . $host_view->hardware->cpuInfo->hz * $host_view->hardware->cpuInfo->numCpuCores . "\n";
	} else {
	    print "Cpu Speed:\tNot Known\n";
	}
	if (defined ($host_view->summary->quickStats->overallCpuUsage)) {
	    print "Cpu Usage:\t" . $host_view->summary->quickStats->overallCpuUsage . "\n";
	} else {
	    print "Cpu Usage:\tNot Known\n";
	}
	if (defined ($host_view->summary->quickStats->overallMemoryUsage)) {
	    print "Memory Usage:\t" . $host_view->summary->quickStats->overallMemoryUsage . "\n";
	} else {
	    print "Memory Usage:\tNot Known\n";
	}
	if (defined ($host_view->summary->rebootRequired)) {
	    print "Reboot Required:\t" . $host_view->summary->rebootRequired . "\n";
	} else {
	    print "Reboot Required:\tNot Known\n";
	}
	if (defined ($host_view->runtime->inMaintenanceMode)) {
	    print "Maintenance mode:\t" . $host_view->runtime->inMaintenanceMode . "\n";
	} else {
	    print "Maintenance mode:\tNot Known\n";
	}
        my $pendingAlarms = "No";
        eval {
            my $alarmState = $host_view->triggeredAlarmState();
            foreach (@$alarmState) {
		if ($_->overallStatus->val eq "red" || $_->overallStatus->val eq "yellow") {
		    $pendingAlarms = "Yes";
		}
            }
        };
        print "Pending Alarms:\t" . $pendingAlarms. "\n";
	print "Model:\t" . $host_view->hardware->systemInfo->model . "\n";
	print "Vendor:\t" . $host_view->hardware->systemInfo->vendor . "\n";
	print "VMotion:\t" . $host_view->summary->config->vmotionEnabled . "\n";
	print "Physical Memory:\t" . $host_view->summary->hardware->memorySize . " bytes\n";

	if (defined ($host_view->summary->config->product)) {
	    print "Software on host:\t" . ${$host_view->summary->config->product}{'fullName'} . "\n";
	}
    }
}


#
# Print information about the VM
#
sub vmInfo($)
{
    my ($vmname) = @_;

    my %toolsStatus = (
		       'toolsNotInstalled' => 'VMware Tools has never been installed or has '
		       .'not run in the virtual machine.',
		       'toolsNotRunning' => 'VMware Tools is not running.',
		       'toolsOk' => 'VMware Tools is running and the version is current',
		       'toolsOld' => 'VMware Tools is running, but the version is not current',
		       );

    my $vmlist;
    die "No VM name specified" unless ($vmname);
    eval {
	$vmlist = Vim::find_entity_views(view_type => 'VirtualMachine',
					 filter => { name => $vmname } ,
					 properties => [ "config", "summary", "guest" ] );
    };
    die $@ if ($@);

    foreach (@$vmlist) {
	my $vm_view = $_;

	if (defined($vm_view->config) && defined ($vm_view->config->name)) {
	    print "Name:\t" . $vm_view->config->name . "\n";
	} else {
	    print "Name:\tNot Known\n";
	}
	if (defined ($vm_view->summary->config->numCpu)) {
	    print "No. of CPU(s):\t" . $vm_view->summary->config->numCpu . "\n";
	} else {
	    print "No. of CPU(s):\tNot Known\n";
	}
	if (defined ($vm_view->summary->config->memorySizeMB)) {
	    print "Memory Size:\t" . $vm_view->summary->config->memorySizeMB . "\n";
	} else {
	    print "Memory Size:\tNot Known\n";
	}
	if (defined ($vm_view->summary->config->numVirtualDisks)) {
	    print "Virtual Disks:\t" . $vm_view->summary->config->numVirtualDisks . "\n";
	} else {
	    print "Virtual Disks:\tNot Known\n";
	}
	if (defined ($vm_view->summary->config->template)) {
	    print "Template:\t" . $vm_view->summary->config->template . "\n";
	} else {
	    print "Template:\tNot Known\n";
	}
	if (defined ($vm_view->summary->config->vmPathName)) {
	    print "vmPathName:\t" . $vm_view->summary->config->vmPathName . "\n";
	} else {
	    print "vmPathName:\tNot Known\n";
	}
	if (defined ($vm_view->summary->runtime->powerState->val)) {
	    print "Power State:\t" . $vm_view->summary->runtime->powerState->val . "\n";
	} else {
	    print "Power State:\tNot Known\n";
	}
	if (defined ($vm_view->summary->guest->guestFullName)) {
	    print "Guest OS:\t" . $vm_view->summary->guest->guestFullName . "\n";
	} else {
	    print "Guest OS:\tNot Known\n";
	}
	if (defined ($vm_view->summary->guest->guestId)) {
	    print "guestId:\t" . $vm_view->summary->guest->guestId . "\n";
	} else {
	    print "guestId:\tNot Known\n";
	}
	if (defined ($vm_view->summary->guest->hostName)) {
	    print "Host name:\t" . $vm_view->summary->guest->hostName . "\n";
	} else {
	    print "Host name:\tNot Known\n";
	}
	if (defined ($vm_view->summary->guest->ipAddress)) {
	    print "IP Address:\t" . $vm_view->summary->guest->ipAddress . "\n";
	} else {
	    print "IP Address:\tNot Known\n";
	}
	if (defined ($vm_view->summary->guest->toolsStatus)) {
	    my $status = $vm_view->summary->guest->toolsStatus->val;
	    print "VMware Tools:\t" . $toolsStatus{$status} . "\n";
	} else {
	    print "VMware Tools:\tNot Known\n";
	}
	if (defined ($vm_view->summary->quickStats->overallCpuUsage)) {
	    print "Cpu usage:\t" . $vm_view->summary->quickStats->overallCpuUsage . " MHz\n";
	} else {
	    print "Cpu usage:\tNot Known\n";
	}
	if (defined ($vm_view->summary->quickStats->hostMemoryUsage)) {
	    print "Host memory usage:\t" . $vm_view->summary->quickStats->hostMemoryUsage . " MB\n";
	} else {
	    print "Host memory usage:\tNot Known\n";
	}
	if (defined ($vm_view->summary->quickStats->guestMemoryUsage)) {
	    print "Guest memory usage:\t" . $vm_view->summary->quickStats->guestMemoryUsage . " MB\n";
	} else {
	    print "Guest memory usage:\tNot Known\n";
	}
	# The caller simply needs to count the number of times IP Addresses are printed and populate their 
	# IP array for the VM based on that
	if (defined ($vm_view->guest->net)) {
        my @guest_net_array = $vm_view->guest->net;
        my @guest_net_settings = @{$guest_net_array[0]};

        foreach my $guest_net_element (@guest_net_settings) {
            print "IP Addresses:\t" . $guest_net_element->{'ipAddress'}[1] . "\n";
        }
    } else {
        print "IP Addresses:\tNot known\n";
    }

    }
}


#
# Print information about vDisks in a VM
#
sub vdiskInfo($)
{
    my ($vmname) = @_;

    my $vmlist;
    die "No VM name specified" unless ($vmname);
    eval {
	$vmlist = Vim::find_entity_views(view_type => 'VirtualMachine',
					 filter => { name => $vmname } ,
					 properties => [ "name", "config", "summary", "runtime" ] );
    };
    die $@ if ($@);

    foreach (@$vmlist) {
	my $vm_view = $_;
    
	my $mor_host = $_->runtime->host;
	my $host = Vim::get_view(mo_ref => $mor_host);

	if (defined($vm_view->config) && defined ($vm_view->config->name)) {
	    print "Name : " . $vm_view->config->name . "\n";
	}

	if (defined $vm_view->config->hardware->device) {
	    my $cnt = 0;
	    my $devices = $vm_view->config->hardware->device;
	    foreach my $dev (@$devices) {
		next unless $dev->isa("VirtualDisk");
		my $backing = $dev->backing;
		next unless ($backing->isa("VirtualDiskFlatVer2BackingInfo") ||
			     $backing->isa("VirtualDiskRawDiskMappingVer1BackingInfo") ||
			     $backing->isa("VirtualDiskPartitionedRawDiskVer2BackingInfo"));
		Util::trace(0, $vm_view->name . " Disk[". $cnt ."]: Name : "
			    . $dev->deviceInfo->label . "\n");
		if ($backing->isa("VirtualDiskRawDiskMappingVer1BackingInfo") ||
		    $backing->isa("VirtualDiskPartitionedRawDiskVer2BackingInfo")) {

		    if (defined($backing->compatibilityMode) && $backing->compatibilityMode eq "physicalMode") {
			# pRDM disk
			# check if this is cache device
			my $lunUUID = $backing->lunUuid;
			my $scsiLun = $host->config->storageDevice->scsiLun;
			my $dtype = "pRDM";
			my $canName = '';
			my $lun;
			foreach (@$scsiLun) {
			    if ($_->lunType eq "disk" && $_->uuid eq $lunUUID) {
				$lun = $_;
				if ($_->model eq "Flash Accel     ") {
				    $dtype = "CACHE";
				}
				$canName = $_->canonicalName;
			    }
			}

			Util::trace(0, $vm_view->name . " Disk[". $cnt ."]: Type : "
				    . $dtype . "\n");
			Util::trace(0, $vm_view->name . " Disk[". $cnt ."]: Id : "
				    . $canName . "\n");
			if (defined($lun)) {
			    Util::trace(0, $vm_view->name . " Disk[". $cnt ."]: Blocks : "
					. $lun->capacity->block . "\n");
			      Util::trace(0, $vm_view->name . " Disk[". $cnt ."]: Blocksize : "
					  . $lun->capacity->blockSize . "\n");
			  } else {
			      my $blocksize = 512;
			      my $blocks = ($dev->capacityInKB * 1024) / 512;
			      Util::trace(0, $vm_view->name . " Disk[". $cnt ."]: Blocks : "
					  . $blocks . "\n");
			      Util::trace(0, $vm_view->name . " Disk[". $cnt ."]: Blocksize : "
					  . $blocksize . "\n");
			  }
		    } else {
			# vRDM disk
			my $uuid = $backing->uuid;
			$uuid =~ s/-//g;
			Util::trace(0, $vm_view->name . " Disk[". $cnt ."]: Type : vRDM\n");
			Util::trace(0, $vm_view->name . " Disk[". $cnt ."]: Id : naa."
				    . $uuid . "\n");
		    }
		}
		if ($backing->isa("VirtualDiskFlatVer2BackingInfo")) {
		    my $uuid = $backing->uuid;
		    $uuid =~ s/-//g;
		    Util::trace(0, $vm_view->name . " Disk[". $cnt ."]: Type : VMDK\n");
		    Util::trace(0, $vm_view->name . " Disk[". $cnt ."]: Id : naa."
				. $uuid . "\n");
		}
		Util::trace(0, $vm_view->name . " Disk[". $cnt ."]: Capacity : "
			    . ($dev->capacityInKB * 1024) . "\n");
		Util::trace(0, $vm_view->name ." Disk[". $cnt ."]: Path : "
			    ."Not Known\n");
		Util::trace(0, $vm_view->name ." Disk[".$cnt."]: Freespace : "
			    ."Not Known\n");
		$cnt++;
	    }
	}
    }
}

#
# ======= MAIN =============
#
my $operation = Opts::get_option('operation');
my $datacenter = Opts::get_option('datacenter');
my $hostname = Opts::get_option('hostname');
my $vmname = Opts::get_option('vmname');

switch ($operation) {
    case "dclist" {
	listDC();
    }
    case "hostlist" {
	listHosts($datacenter);
    }
    case "vmlist" {
	listHostsVMs($datacenter);
    }
    case "hostinfo" {
	hostInfo($hostname);
    }
    case "maintmode" {
	getMaintenanceMode($hostname);
    }
    case "vminfo" {
	vmInfo($vmname);
    }
	case "restartSfcbdWatchdog" {
	serviceSfcbdWatchdogRestart($hostname);
	}
	case "serviceGetSfcbdWatchdogStatus" {
	serviceGetSfcbdWatchdogStatus($hostname);
	}
    case "vdiskinfo" {
	vdiskInfo($vmname);
    }
    else {
	die "Invalid operation '$operation'";
    }
}

Util::disconnect();
