#!/usr/bin/perl -w

use strict;
use warnings;
use POSIX;

use VMware::VIRuntime;
use VMware::VILib;
use VMware::VICredStore;

# misc sizes
use constant KB => 1024;
use constant MB => 1024 * 1024;
use constant GB => 1024 * 1024 * 1024;

# Constants for the range of Controller Keys
use constant VM_SCSI_CONTROLLER_LO      => 1000;
use constant VM_SCSI_CONTROLLER_HI      => 1003;
use constant VM_IDE_CONTROLLER_LO       => 200;
use constant VM_IDE_CONTROLLER_HI       => 201;
use constant VM_IDE_DISK_LO             => 3000;
use constant VM_IDE_DISK_HI             => 3003;
use constant VM_NETWORK_LO              => 4000;

use constant VM_UNUSED_SCSI_PORT        => 7;
use constant VM_MAX_SCSI_CTRLS          => 4;
use constant VM_MAX_SCSI_PORTS          => 16;

my %opts = (
     hostname => {
     type => "=s",
     variable => "HOSTNAME",
     help => "Name of managed ESX host",
     required => 0,
     },
     vmname => {
     type => "=s",
     variable => "VMNAME",
     help => "Name of Virtual Machine",
     required => 0,
     },
     cachespace => {
     type => "=s",
     variable => "CacheSpace",
     help => "Name of cachespace to attach to",
     required => 0,
     },
     datastore => {
     type => "=s",
     variable => "Datastore",
     help => "Name of datastore root",
     required => 0,
     },
	 diskCanonicalName => {
     type => "=s",
     variable => "diskCanonicalName",
     help => "Canonical name associated with the disk to be claimed",
     required => 0,
     },
     operation => {
     type => "=s",
     variable => "operation",
     help => "Operation can be create or remove disk",
     required => 1,
     }

);
Opts::add_options(%opts);
Opts::parse();
Opts::validate();

# Given a VM name, find the VM object
# Args:
#   $vmName - VM Name
# Returns:
#   vm (view) - success
#   (undefined) - error
#
sub findVM($) {
    my ($vmName) = @_;

    # if vmName was not specified, return
    if (not defined $vmName) { return; }

    return Vim::find_entity_view(view_type => "VirtualMachine",
                                 filter => { "summary.config.name" => $vmName });
}

#
# Find a VMFS datastore on the host
# Args:
#   $host - HostSystem
# Returns:
#   ds - datastore object
#
sub findVMFS {
    my ($host) = @_;
    my $refreshed = 0;

    # validate host argument
    if (! $host->isa("HostSystem")) {
	die "Error: Invalid HostSystem";
    }

REFRESH2:
    my $datastores;
    eval {
        $datastores = Vim::get_views(mo_ref_array => $host->datastore);
    };
    if ($@) {
        if ($refreshed) { return; }
        $host->update_view_data();
        $refreshed = 1;
        goto REFRESH2;
    }

    foreach(@$datastores) {
        if ($_->summary->type eq 'VMFS') {
	    print ("Using VMFS datastore '" . $_->summary->name . "' for RDM disk\n");
            return $_;
        }
    }
    return;
}
	

#
# Given a datastore name, find the datastore object
# Args:
#   $host - HostSystem
#   $dsName - Datastore name
# Returns:
#   ds (view) - success
#   (undefined) - error
#
sub findDS {

    my ($host, $dsName) = @_;
    my $refreshed = 0;

    # validate host argument
    if (! $host->isa("HostSystem")) {
	die "Error: Invalid HostSystem";
    }

    # if dsName was not specified, return^M
    if (not defined $dsName) {
	die "Error: Missing datastore name";
    }

REFRESH1:
    my $datastores;
    eval {
        $datastores = Vim::get_views(mo_ref_array => $host->datastore);
    };
    if ($@) {
        if ($refreshed) { return; }
        $host->update_view_data();
        $refreshed = 1;
        goto REFRESH1;
    }

    foreach(@$datastores) {
        if ($_->summary->name eq $dsName) {
            return $_;
        }
    }

    return;
}


# Given a SCSI disk canonical name, find the HostScsiDisk object
# Args:
#   $host - HostSystem
#   $diskName - Disk canonical name
# Returns:
#   scsidisk (view) - success (HostScsiDisk object)
#   (undefined) - error
#
sub findDiskByName {

    my ($host, $diskName) = @_;

    # validate host argument
    if (! $host->isa("HostSystem")) {
        die "Error: Invalid HostSystem";
    }

    # if diskName was not specified, return
    if (not defined $diskName) {
        die "Error: Missing disk name";
    }

    my $hostname = $host->summary->config->name;
    my $notfound = 60;
    while ($notfound--) {
	my $scsiLun = $host->config->storageDevice->scsiLun;
	foreach (@$scsiLun) {
	    if ($_->lunType eq "disk" && $_->canonicalName eq $diskName) {
		return $_;
	    }
	}
	select(undef, undef, undef, 0.25);      # sleep 0.25 secs

	# refetch the host
	my %args = ( view_type => "HostSystem",
		     filter => { name => $hostname } );
	$host = Vim::find_entity_view(%args);
    }

    return;
}

# Given a host, find the Flash Accel disks that aren't vRDM mapped
# Args:
#   $host - HostSystem
# Returns:
#	None
#
sub getOrphanedCacheSpaces {
    my ($host) = @_;
    my %flashAccelCacheSpaces = ();
    my @vmMappedRDMDisks;

    #validate host argument
    if (! $host->isa("HostSystem")) {
        die "Error: Invalid HostSystem";
    }

    my $scsiLun = $host->config->storageDevice->scsiLun;

    foreach (@$scsiLun) {
        my $lun = $_;
        my $lunType = $lun->lunType;
        next if ($lunType ne "disk");

        my $canName = $lun->canonicalName;
        my $vendor = $lun->vendor;
        my $product = $lun->model;
        my $uuid = $lun->uuid;
        next if ($vendor !~ /NetApp/i || $product !~ /Flash\ Accel/i);

        $flashAccelCacheSpaces{$lun->uuid} = $lun->canonicalName;
    }

	my $noFlashAccelCacheSpaces = scalar keys %flashAccelCacheSpaces;
	if ($noFlashAccelCacheSpaces == 0) { 
		return;
	}

	my $vms;
    eval { $vms = Vim::get_views(mo_ref_array => $host->vm,
					properties => [ "config" ]); };
    if (not defined $vms) { goto printOrphanedLuns};

    foreach my $vm (@$vms) {
        my $devices = $vm->config->hardware->device;
        foreach my $device (@$devices) {
            if ($device->isa("VirtualDisk") &&
                ( $device->backing->isa("VirtualDiskRawDiskMappingVer1BackingInfo") ||
                  $device->backing->isa("VirtualDiskPartitionedRawDiskVer2BackingInfo") )) {
                my $devname = $device->backing->deviceName;
                $devname =~ s/vml\.//g;
				if (defined $flashAccelCacheSpaces{$devname}) {
					delete($flashAccelCacheSpaces{$devname});
				}
            }
        }
    }

printOrphanedLuns:
	for my $key (keys %flashAccelCacheSpaces) {
        my $value = $flashAccelCacheSpaces{$key};
        print "NAAID:\t".$value. "\n";
    }
}


# Given a SCSI disk canonical name, find the VM and vDisk that points to this disk
# Args:
#   $host - HostSystem
#   $diskName - Disk canonical name
# Returns:
#   $vm - VM object
#   $vdisk - virtual disk object
#
sub findVDiskByOwner {
    my ($host, $diskName) = @_;

    # validate host argument
    if (! $host->isa("HostSystem")) {
        die "Error: Invalid HostSystem";
    }

    # if diskName was not specified, return
    if (not defined $diskName) {
        die "Error: Missing disk name";
    }

    my $disk = findDiskByName($host, $diskName);
    if (not defined $disk) {
	die "Error: No such disk";
    }

    my $vms;
    eval { $vms = Vim::get_views(mo_ref_array => $host->vm); };
    return if ($@);
    return if (not defined $vms);

    foreach my $vm (@$vms) {
  	my $devices = $vm->config->hardware->device;
        foreach my $device (@$devices) {
            if ($device->isa("VirtualDisk")) {
            	if ($device->backing->isa("VirtualDiskRawDiskMappingVer1BackingInfo") ||
                    $device->backing->isa("VirtualDiskPartitionedRawDiskVer2BackingInfo")) {
		    my $devname = $device->backing->deviceName;
		    $devname =~ s/vml\.//g;
		    if ($devname eq $disk->uuid) {
			return ($vm, $device);
		    }
		}
	    }
	}
    }
}


#
# Wait for a task to complete
# Args:
#   task - task we are waiting to complete
# Returns:
#   nothing
#
sub waitForTask {
    my ($task) = @_;
    my $state;
    my $lastState = "";

    while (1) {
        $state = $task->info->state->val;
        last if $state eq "success" || $state eq "error";
# IF DEBUG
#       if ( defined $task->info->progress ) {
#           $state = $state . " " . $task->info->progress . "%";
#       }
#       print "\r\t$state" unless $state eq $lastState;
#       $lastState = $state;
# ENDIF
        select(undef, undef, undef, 0.25);      # sleep 0.25 secs
        $task->update_view_data();
    }

    if ( $task->info->state->val eq "error" ) {

          # This task has resulted in an error

          # Special-case the TaskInProgress fault, dying w/TaskInProgressErr
          my $taskInProgressErr = taskInProgressErr($task->info->error->fault, $task);
          die $taskInProgressErr if defined $taskInProgressErr;

          # Otherwise generate and throw a SOAP fault.
          my $faultMsgs = $task->info->error->fault->faultMessage;
          my $errMsgs = $task->info->error->localizedMessage . "\n";
          foreach my $fmsg (@$faultMsgs) {
              $errMsgs .= $fmsg->message . "\n";
          }

          my $soapFault = SoapFault->new;
          $soapFault->detail($task->info->error->fault);
          $soapFault->fault_string($errMsgs);
          die $soapFault;
    }
}

# Transform a VMware TaskInProgress fault to our TaskInProgressErr
# Args:
#   $err - an error object
#   $failedTask - task which failed
# Returns:
#   new TaskInProgressErr if $err is a VMware TaskInProgress fault
#   undef otherwise
#
sub taskInProgressErr($;$) {
    my ($err, $failedTask) = @_;
    return unless (defined $err && ref($err));
    $err = $err->detail if $err->isa("SoapFault");
    return unless $err->isa("TaskInProgress");

    my $task = Vim::get_view(mo_ref => $err->task);
    my $details = $failedTask->info->descriptionId . " failed due to " .
        $task->info->descriptionId . " from ";
    my $reason = $task->info->reason;
    if ($reason->isa("TaskReasonUser")) {
        $details .= "user " . $reason->userName;
    } elsif ($reason->isa("TaskReasonSchedule")) {
        $details .= "scheduled task " . $reason->name;
    } elsif ($reason->isa("TaskReasonAlarm")) {
        $details .= "alarm " . $reason->alarmName . " on " . $reason->entityName;
    } else {
        # TaskReasonSystem (no details) and any unknowns
        $details .= ref($reason);
    }
    # The progress and startTime methods may not exist (or may return
    #   undef), so include this info when it exists and otherwise
    #   ignore errors accessing it.
    eval {
        my $progress = $task->info->progress;
        $details .= "  " . $progress . "% done" if defined $progress;
    };
    eval {
        my $startTime = $task->info->startTime;
        $details .= "  started " . $startTime if defined $startTime;
    };
    return new TaskInProgressErr($details);
}

#
# Check RdmFilter.HbaIsShared option
# Args:
#   $host - Host System
#
# Returns:
#   nothing
sub testAndSetRdmFilterHbaIsShared {
    my ($host) = @_;
    my @getQueryArray;
	my $false = PrimType->new(0, 'boolean');

    my $optionManager = Vim::get_view(mo_ref => $host->configManager->advancedOption);

    eval {
    @getQueryArray = $optionManager->QueryOptions(name => 'RdmFilter.HbaIsShared');
    };
    if ($@) {
        die "Error: Could not Query RdmFilter Option";
    };

    my $val = $getQueryArray[0][0]->value;

    if ($val eq "true") {
        my $changeValue = new OptionValue (key => "RdmFilter.HbaIsShared", value => $false);

        eval {
            $optionManager->UpdateOptions(changedValue => [$changeValue]);
        };
        if ($@) {
            die "Error: Could not set RdmFilter.HbaIsShared Option to false";
        };

    };

    print $val;
}

#
# Given a datastore name, find the datastore object
# Args:
#   $host - HostSystem
#   $dsName - Datastore name
# Returns:
#   ds (view) - success
#   (undefined) - error
#
sub dsfind {

    my ($host, $dsName) = @_;
    my $refreshed = 0;

    # validate host argument
    if (! $host->isa("HostSystem")) {
	die "Error: Invalid HostSystem";
    }

    # if dsName was not specified, return
    if (not defined $dsName) {
	die "Error: Missing datastore name";
    }

REFRESH1:
    my $datastores;
    eval {
	$datastores = Vim::get_views(mo_ref_array => $host->datastore);
    };
    if ($@) {
	if ($refreshed) { return; }
	$host->update_view_data();
	$refreshed = 1;
	goto REFRESH1;
    }

    foreach(@$datastores) {
        if ($_->summary->name eq $dsName) {
	    return $_;
        }
    }

    return;
}

#
# Given a datastore name, find the associated datacenter
# Args:
#   $store - Datastore name
# Returns:
#   datacenter (view) - success
#   (undefined) - error
#
sub findDC {

    my $store = $_[0];
    my $datactr;

    my $datacenters = Vim::find_entity_views(view_type => "Datacenter");
DC: foreach my $dc (@$datacenters) {
        my $datastores = $dc->datastore;
	foreach my $dsref (@$datastores) {
	    my $ds = Vim::get_view(mo_ref => $dsref);
	    if ( $ds->info->name eq $store ) {
		$datactr = $dc;
		last DC;
	    }
	}
    }

    return $datactr;
}

#
# Create Directory in Datastore
# Args:
#   dirName - directory name
#   datastore - datastore object where the directory will be created
# Returns:
#   Name of directory created (or which already exists)
#
sub createDir {
    my ($dirName, $datastore) = @_;

    my $store = $datastore->info->name;
    my $datactr = findDC($store)
	or die "Error: cannot find Datacenter for pool '$store'";

    my $storeDir = "[$store] $dirName";

    my $browser = Vim::get_view(mo_ref => $datastore->browser);

    eval {
        my $task = $browser->SearchDatastore_Task(datastorePath => $storeDir);
        waitForTask(Vim::get_view(mo_ref => $task));
    };
    if ($@) {
        if (ref($@) ne 'SoapFault') { return; }
        if (ref($@->detail) eq "FileNotFound") {
            eval {
                my $fileMgr = Vim::get_service_content()->fileManager;
                $fileMgr = Vim::get_view(mo_ref => $fileMgr);
                $fileMgr->MakeDirectory(name => $storeDir,
                                        datacenter => $datactr,
                                        createParentDirectories => 1);
            };
            if ($@) { return; }
        } else {
            return;
        }
    }

    return $storeDir;
}

# Internal.  Return a new filename in the given datastore directory,
# which must exist.
# Args:
#    datastore - the Datastore object
#    storedir  - existing directory (path) within datastore
#    vmName    - VM name
#    skip      - number of available names to skip before returning next
# Returns
#    unused filename under storedir
#
sub newBackingFilename {
    my ($datastore, $storeDir, $vmName, $skip) = @_;
    my $task;
    my %vmdkFiles = ();

    # Build hash of VM vmdk filenames in datastore directory and then
    #   choose a backing file name that doesn't collide.
    my $dsBrowser = Vim::get_view(mo_ref => $datastore->browser);
    my $dsName = $datastore->info->name;

    eval {
	$task = $dsBrowser->SearchDatastore_Task(datastorePath => $storeDir);
	$task = Vim::get_view(mo_ref => $task);
	waitForTask($task);
    };
    if ($@) {
	die "Error: Search pool '$dsName' dir '$storeDir' failed";
    };

    my $files = $task->info->result->file;
    for my $file (@$files) {
	my $path = $file->path;
	if ($path =~ /\.vmdk$/) {
	    $path =~ s|.*/||;		# strip [store] directory
	    $vmdkFiles{$path} = 1;
	}
    }
    my $fpath;
    my $filenum = 1;
    do {
	$fpath = $vmName . "_" . $filenum++ . ".vmdk";
    } while (defined $vmdkFiles{$fpath} || $skip--);

    $vmdkFiles{$fpath} = 1;
    return $fpath;
}


# Returns the next available "slot" (controllerKey, unitNum) for a SCSI
# disk in the given VM.
# Args:
#   vm - the VirtualMachine
#   skip - number of available slots to skip before returning next
# Returns:
#   (controllerKey, unitNum) for success
#   undef on failure
# Side effects:
#   Creates a SCSI controller corresponding to the returned controllerKey
#   if it doesn't already exist.
#
sub findScsiSlot {

    my ($vm, $startCtrl, $startUnit, $skip) = @_;
    my @scsiDisks = ([], [], [], []);
    my $devs = $vm->config->hardware->device;

    # build a table of free and used scsi ports
    for my $dev (@$devs) {
	if ( defined $dev->controllerKey &&
	     $dev->controllerKey >= VM_SCSI_CONTROLLER_LO &&
	     $dev->controllerKey <= VM_SCSI_CONTROLLER_HI ) {
	    $scsiDisks[$dev->controllerKey -
		       VM_SCSI_CONTROLLER_LO][$dev->unitNumber] = 1;
	}
    }

    my $controllerNum = (defined $startCtrl) ? $startCtrl : 0;
    my $unitNum = (defined $startUnit) ? $startUnit : 0;

    # find an available port for this disk. We will first allocate all
    # the ports on a given controller before moving to the next controller.
    while ($controllerNum < VM_MAX_SCSI_CTRLS) {

	if (! (defined($scsiDisks[$controllerNum][$unitNum]) || $skip--)) {
	    last;
	}

	print ("SCSI slot [$controllerNum][$unitNum] is in use\n");

	$unitNum++;
	# skip unused SCSI port
	$unitNum++ if ($unitNum == VM_UNUSED_SCSI_PORT);
	if ($unitNum >= VM_MAX_SCSI_PORTS) {
	    $unitNum = 0;
	    $controllerNum++;
	}
    }
    return undef if ($controllerNum >= VM_MAX_SCSI_CTRLS);

    # make sure controller exists.  unused controllers are sometimes removed.
    my $controllerKey = $controllerNum + VM_SCSI_CONTROLLER_LO;
    my $dev = findVirtDevice($vm, $controllerKey);
    if (not $dev) {
	my $sharing = new VirtualSCSISharing("noSharing");
	my $controller = new VirtualLsiLogicSASController(busNumber => $controllerNum,
							  controllerKey => 100,
							  key => $controllerKey,
							  unitNumber => $controllerNum + 3,
							  sharedBus => $sharing);
	my $op = new VirtualDeviceConfigSpecOperation("add");
	my $vdcs = new VirtualDeviceConfigSpec(device => $controller,
					       operation => $op);
	my $vmcs = new VirtualMachineConfigSpec(deviceChange => [ $vdcs ]);
	eval {
	    my $task = Vim::get_view(mo_ref => $vm->ReconfigVM_Task(spec => $vmcs));
	    waitForTask($task);
	};
	if ($@) {
	    die "Error: could not create new SCSI controller $controllerNum";
	}
	$vm->update_view_data();
    }

    $scsiDisks[$controllerNum][$unitNum] = 1;
    return ($controllerNum, $unitNum);
}

#
# Create a virtual disk
# Args:
#   vm - Virtual machine object
#   controllerKey - Controller number for virtual disk
#   unitNum - Unit number for virtual disk
#   sizeInKB - Size of virtual disk in kilobytes
#   backing - Description of the datastore backing this virtual disk
# Returns:
#   nothing
#
sub createVirtualDisk {
    my ($vm, $controllerKey, $unitNum, $sizeInKb, $backing) = @_;

    #
    # NASTY ESX BUG: Be *sure* to specify unitNumber below, otherwise
    #  the reconfig task runs forever and can't be canceled, leaving the
    #  VM locked.  Rebooting the ESX host seems to be the only recourse.
    #
    print ("Create virtual disk $controllerKey:$unitNum size=$sizeInKb\n");

    my $vdisk = new VirtualDisk(backing => $backing,
                                key => 9999,    # ignored
                                controllerKey => $controllerKey,
                                capacityInKB => $sizeInKb,
                                unitNumber => $unitNum);

    my $op = new VirtualDeviceConfigSpecOperation("add");
    my $fileOp = new VirtualDeviceConfigSpecFileOperation("create");
    my $spec = new VirtualDeviceConfigSpec(device => $vdisk,
                                           fileOperation => $fileOp,
                                           operation => $op);
    my @vdcs;
    push(@vdcs, $spec);

    # Perform the actual reconfiguration task: create and add the disks.
    # We use the asynchronous '_Task' version here so we can monitor progress.
    #
    my $task;
    my $retry = 1;
    my $reconfig_spec = new VirtualMachineConfigSpec(deviceChange => \@vdcs);

    my $rspec = $reconfig_spec;
RETRY_CREATE:
    eval {
        #Util::maybeInject("vm_disk_create_err");
        $task = $vm->ReconfigVM_Task(spec => $reconfig_spec);
        waitForTask(Vim::get_view(mo_ref => $task));
    };
    if ($@) {
        if ($@->isa("TaskInProgressErr")) {
            if ($retry) {
                print("retrying disk create @ $controllerKey:$unitNum size=$sizeInKb\n");
                $retry = 0;
                $reconfig_spec = $rspec;
                select(undef, undef, undef, 1); # sleep 1 sec
                goto RETRY_CREATE;
            }
            die $@;
        }
    }
}


# Lookup a virtual device by (numeric) device key
# Args:
#   $vm  - VirtualMachine whose device we're looking up
#   $key - device key
# Returns:
#   the device object, or undef if no such device
#
sub findVirtDevice($$) {
    my ($vm, $key) = @_;
    my $devs = $vm->config->hardware->device;
    foreach (@$devs) {
        if ($_->key eq $key) { return $_; }
    }
    return;
}

# Subroutine to check if the input datastore is of type VMFS
# Args:
# host      -   Physical host on which datastore exists
# datastore -   The one for which we check if it is of type VMFS
# Returns:
#	nothing:
sub checkIfVMFSDataStore {
    my ($host, $canonicalName) = @_;

    my $mountInfo = $host->config->fileSystemVolume->mountInfo;
    # check to see if this device is mounted
    for (@$mountInfo) {
        my $volume = $_->volume;
        if ($volume->isa("HostVmfsVolume")) {
            my $extents = $volume->extent;
            my $volname = $volume->name;
            for (@$extents) {
                if ($_->diskName eq $canonicalName) {
                    print ("Error: Disk ".$canonicalName." is associated with a VMFS volume ".$volname.".\n");
                }
            }
        }
    }
}

# Subroutine for checking if a VMFS datastore exists for RDM mapping
# Args:
# host		-	Physical host on which VM is hosted
# datastore -	Optional argument specifying a particular datastore to look for
# Returns:
#	nothing on success
#	'Error' message on failure
sub checkForVMFSDataStore {
	my ($host, $datastore, $vm) = @_;

	# find the datastore VM is using if a datastore was not specified
	if (!$datastore) {
    my $dsName = $vm->config->files->vmPathName;
    $dsName =~ s/\].*//;
    $dsName =~ s/\[//;    # extract the dsName
    $datastore = dsfind($host, $dsName)
        or die "Error: could not find datastore '$dsName'";
    }
	
	# If the datastore is not a VMFS, find a VMFS DS
	if ($datastore->summary->type ne "VMFS") {
    my $newDS = findVMFS($host);
    $datastore = $newDS if ($newDS);
    }

    my $dsName = $datastore->summary->name;
    die "Error: Backing datastore '$dsName' is not type VMFS"
    unless ($datastore->summary->type eq "VMFS");

	print ("VMFS datastore=$dsName\n");
}
	

# Subroutine for create a RDM disk in the VM
# Prerequisites:
#	It is assumed that the checkForVMFSDataStore function is run before calling into this 
#	routine. The datastore presented must be a valid VMFS datastore on which RDM mappings
#	can be stored.
# Args:
#   vm         - Virtual machine object
#   device     - Physical disk (HostScsiDisk)
#   mode       - Compatibility mode of RDM disk - virtual, physical
# Returns:
#   nothing
sub createRdmDisk {
    my ($host, $vm, $mode, $disk, $datastore) = @_;
    my $vmName = $vm->summary->config->name;
	
	my $dsName = $datastore->summary->name;
	die "Error: Backing datastore '$dsName' is not type VMFS"
	unless ($datastore->summary->type eq "VMFS");

    # get the disk size
    my $sizeInKb =
	($disk->capacity->block * $disk->capacity->blockSize) / KB;
    print ("Disk size = $sizeInKb KB\n");

    # find a unused scsi slot in the VM
    my ($controllerNum, $unitNum) = findScsiSlot($vm);
    if (not defined $controllerNum) {
	die "Error: No SCSI ports available for this disk.";
    }
    print ("Using SCSI slot c=$controllerNum,u=$unitNum\n");

    my $storeDir = createDir($vmName, $datastore)
	or die "Error: Cannot create datastore directory";
    my $fpath = newBackingFilename($datastore, $storeDir, $vmName);

    # Create a RDM backing
    my $backing = new VirtualDiskRawDiskMappingVer1BackingInfo(compatibilityMode => $mode,
							       diskMode => "independent_persistent",
							       deviceName => $disk->deviceName,
							       fileName => "$storeDir/$fpath");
    eval {
	createVirtualDisk($vm,
			  $controllerNum + VM_SCSI_CONTROLLER_LO,
			  $unitNum,
			  $sizeInKb,
			  $backing);
    };
    if ($@) {
	die "Error: Failed to create virtual disk @ scsi$controllerNum:$unitNum\n" . $@;
    }
    $vm->update_view_data();
    print ("Added $sizeInKb KB disk [$controllerNum:$unitNum] backed by $storeDir/$fpath\n");
    
}


#
# Subroutine for destroying a virtual disk
# Args:
#   vm - Virtual machine
#   disk - VirtualDisk device object
# Returns:
#   nothing
#
sub destroyVirtualDisk {
    my ($vm, $disk) = @_;

    my $fileName = $disk->backing->fileName;

    # ok we found the disk
    # first remove the disk from the vm by doing a vm reconfig
    my @vdcSpecs;
    my $op = new VirtualDeviceConfigSpecOperation("remove");
    my $fileOp = new VirtualDeviceConfigSpecFileOperation("destroy");
    push(@vdcSpecs, new VirtualDeviceConfigSpec(device => $disk,
                                                fileOperation => $fileOp,
                                                operation => $op));
    my $reconfig_spec = new VirtualMachineConfigSpec(deviceChange => \@vdcSpecs);
    eval {
        my $task = $vm->ReconfigVM_Task(spec => $reconfig_spec);
        waitForTask(Vim::get_view(mo_ref => $task));
    };
    if ($@) {
        die $@ if $@->isa("TaskInProgressErr");
        die "Error: Failed to delete disk: $@\n";
    }

    #
    # An ESX bug (SR1573586621) exists when attempting to delete the
    # backing file for a virtual disk attached to a VM which has taken
    # a configuration snapshot (vm config backup) when connected through
    # vCenter (instead of directly to the ESX host.)  Eventhough the disk
    # in question was marked as 'independent', the backing file is not
    # deleted when a "destroy" file operation is used when removing the
    # virtual disk.
    # To work around this issue, we will attempt to delete the backing file
    # directly after removing the virtual disk.  If removing the backing file
    # fails because it doesn't exist, this means that the removing of the
    # virtual disk was successful is also removing the backing file.
    #
    my $service_content = Vim::get_service_content();
    my $apiType = $service_content->about->apiType;
    if ($apiType eq "VirtualCenter") {
        my $dsName = $fileName;
        $dsName =~ s/\[//;
        $dsName =~ s/\].*//;    # strip out datastore name

        # Find the datacenter associated with this datastore.
        my $datactr = findDC($dsName);

        if (!defined $datactr) {
            die "Error: could not find datacenter for datastore '$dsName'";
        }

        my $fm = Vim::get_view (mo_ref => $service_content->{fileManager});

        eval {
            $fm->DeleteDatastoreFile(name => $fileName, datacenter => $datactr);
            $fileName =~ s/\.vmdk/-flat.vmdk/;
            $fm->DeleteDatastoreFile(name => $fileName, datacenter => $datactr);
        };
        if ($@) {
            if (!$@->detail->isa('FileNotFound')) {
                die "Error: Failed to delete backing file $fileName. This file must be deleted manually\n";
            }
        }
    }
	print "Disk remove successful\n";
}

#hostname, vmname, name of cachespace, root of the datastore(must be vmfs)
Util::connect();

my ($hostname, $vmname, $cachespace, $dsName, $operation, $diskCanonicalName);
$hostname = Opts::get_option('hostname');
$vmname = Opts::get_option('vmname');
$cachespace = Opts::get_option('cachespace');
$dsName = Opts::get_option('datastore');
$operation = Opts::get_option('operation');
$diskCanonicalName = Opts::get_option('diskCanonicalName');

my $host = Vim::find_entity_view(
     view_type => 'HostSystem',
     filter => { name => $hostname },
);

unless (defined $host) {
     Util::disconnect();
     die "Error: No host matching '$hostname' found.\n";
}
if ($operation eq "remove") {
    my ($vm, $vdisk) = findVDiskByOwner($host, $cachespace);
	die "Error: Disk '$cachespace' is not assigned to any VM.\n" if ((not defined $vm) || (not defined $vdisk));
	destroyVirtualDisk($vm, $vdisk);
}
elsif ($operation eq "testAndSetRdmFilterHbaIsShared") {
	testAndSetRdmFilterHbaIsShared($host);
}
elsif ($operation eq "checkForVMFSDatastore") {
	my $vm=findVM($vmname) or die "Error: Cannot find VM '$vmname'";
	my $datastore=findDS($host, $dsName) or die "Error: Cannot find datastore '$dsName'";
	checkForVMFSDataStore($host,$datastore,$vm);
}
elsif ($operation eq "checkIfVMFSDatastore") {
    checkIfVMFSDataStore($host,$diskCanonicalName);
}
elsif ($operation eq "getOrphanedCacheSpaces") {
	getOrphanedCacheSpaces($host);
}
else {
	my $mode="physicalMode";
	my $vm=findVM($vmname) or die "Error: Cannot find VM '$vmname'";
        print "Cachespace: " . $cachespace ."\n";
	my $disk=findDiskByName($host, $cachespace) or die "Error: Cannot find disk '$cachespace'";
	my $datastore=findDS($host, $dsName) or die "Error: Cannot find datastore '$dsName'";
	createRdmDisk($host, $vm, $mode, $disk, $datastore);
}

Util::disconnect();

